/* Define iluminacao em um plano. Ocorre GPF se um dos vetores for nulo.
Para funcionar corretamente o poligono deve formar um plano (pelo menos
deve possuir 3 vertices distintos) e pertencer a apenas um plano.
Angulos e graus sao definidos de acordo com NANGS.
Convencoes:
Quanto maior Z menor a iluminacao (cor menor, mais proximo a MINCOR).
Em CIlum::mRetCorZ se Z >= FIMV entao cor eh MINCOR, se Z <= 0 entao cor
eh MAXCOR. */



#include <math.h>
#include <stdlib.h>

#include "defs.h"
#include "Pol.h"

#include "ilum.h"



// Calcula o modulo (comprimento) de um vetor
#define MODULO(X, Y, Z) ( (float) sqrt ( ( X ) * ( X ) + ( Y ) * ( Y ) + ( Z ) * ( Z ) ) )



/* construtor - inicia componentes do vetor de iluminacao com 0
**********************************************************/
CIlum::CIlum ()
{
	m_nIX = m_nIY = m_nIZ = 0;
	m_fModI = 0.0;
}



/* construtor - define valores para vetor de iluminacao
**********************************************************/
CIlum::CIlum ( INT16 nVX, INT16 nVY, INT16 nVZ )
{
	mAtrIlum ( nVX, nVY, nVZ );
}



/* Define componentes do vetor de iluminacao
**********************************************************/
void CIlum::mAtrIlum ( INT16 nVX, INT16 nVY, INT16 nVZ )
{
	m_nIX = nVX;
	m_nIY = nVY;
	m_nIZ = nVZ;
	m_fModI = MODULO ( m_nIX, m_nIY, m_nIZ );
}



/* Define uma cor de acordo com o angulo entre um vetor do pol e
o vetor de iluminacao. return cor (entre MAXCOR e MINCOR)
Usa modelo semelhante ao flat. A diferenca eh que o angulo eh calculado
nao entre o vetor de iluminacao e o vetor normal ao plano e sim entre
um vetor pertencente ao plano do poligono e o vetor de iluminacao.
Com isso poupa-se tempo de calculo do vetor normal. O vetor do plano
pertence ao plano e nao eh normal ao plano.
**********************************************************/
INT16 CIlum::mRetCorP ( CPol &oPol )
{
	// define vetor pertencente ao poligono usando os vertices do poligono
	INT16
		nPX,
		nPY,
		nPZ;
	// verifica se os vertices 0 e 1 estao sobrepostos, se nao estao usa 0 e 1
	if ( oPol.m_anX [ 0 ] != oPol.m_anX [ 1 ] ||
		  oPol.m_anY [ 0 ] != oPol.m_anY [ 1 ] ||
		  oPol.m_anZ [ 0 ] != oPol.m_anZ [ 1 ] ) {

		nPX = oPol.m_anX [ 1 ] - oPol.m_anX [ 0 ];
		nPY = oPol.m_anY [ 1 ] - oPol.m_anY [ 0 ];
		nPZ = oPol.m_anZ [ 1 ] - oPol.m_anZ [ 0 ];
	}
	else {

		nPX = oPol.m_anX [ 2 ] - oPol.m_anX [ 0 ];
		nPY = oPol.m_anY [ 2 ] - oPol.m_anY [ 0 ];
		nPZ = oPol.m_anZ [ 2 ] - oPol.m_anZ [ 0 ];
	}

	// calcula o angulo entre os 2 vetores: iluminacao e poligono
   // define o produto escalar entre os dois vetores
	float
		fProdEsc = nPX * m_nIX + nPY * m_nIY + nPZ * m_nIZ;

   // define o angulo (entre 0 e NANGS2 - 1)
   float
   	fV = (float) fProdEsc / ( MODULO ( nPX, nPY, nPZ ) * m_fModI );
   // verifica erros nos calculos
   if ( fV > 1 )
   	fV = 1;
   else
   	if ( fV < -1 )
      	fV = -1;
	INT16
		nAngulo = (INT16) ( acos ( fV ) * NANGS2 / Pi );

   // ajusta angulo para intervalo de 0 a NANGS / 4     (se angulo maior que
   // Porque:   VV   (iluminacao)                        NANGS2 calcula
   //           \/   (4 paredes)                         complementar)
   //           /\  todas devem receber a mesma iluminacao (sem sombras)
   if ( nAngulo > NANGS4 )
   	nAngulo = NANGS4 - nAngulo + NANGS4;

	// define cor fazendo regra de tres entre angulo e faixa cores utilizaveis
	return ( MINCOR + ( NUMCOR - 1 ) * nAngulo / NANGS4 );
}



/* Define uma cor de acordo com o angulo entre um vetor normal
ao poligono e o vetor de iluminacao. return cor (entre MAXCOR e MINCOR).
Usa modelo flat. Como vetor normal eh obtido:
Seja u e v dois vetores pertencentes ao plano do poligono
(obtidos por vertices nao colinearess do poligono), n o vetor normal,
i, j ,k os componentes de um vetor unitario (iguais a 1).
         | i  j  k|
         |uX uY uZ| = (Ai, Bj, Ck)    => n = (A, B, C)
         |vX vY vZ|   determinante
**********************************************************/
INT16 CIlum::mRetCorN ( CPol &oPol )
{
	// define vetor normal ao plano do poligono
   // Define 2 vetores nao sobrepostos pertencentes ao poligono, o certo
   // seria 2 vetores nao colineares, mas como vertices sao fixos relativo ao
	// poligono o modelador deve tomar cuidado para nao definir 3 ou mais
	// vertices na mesma linha.
	INT16
		nP1X,
		nP1Y,
		nP1Z,
		nP2X,
		nP2Y,
		nP2Z;
	// verifica se os vertices 0 e 1 estao sobrepostos, se nao estao usa 0 e 1
	if ( oPol.m_anX [ 0 ] != oPol.m_anX [ 1 ] ||
		  oPol.m_anY [ 0 ] != oPol.m_anY [ 1 ] ||
		  oPol.m_anZ [ 0 ] != oPol.m_anZ [ 1 ] ) {

		nP1X = oPol.m_anX [ 1 ] - oPol.m_anX [ 0 ];
		nP1Y = oPol.m_anY [ 1 ] - oPol.m_anY [ 0 ];
		nP1Z = oPol.m_anZ [ 1 ] - oPol.m_anZ [ 0 ];

		// verifica se os vertices 0 e 2 estao sobrepostos, se nao usa 0 e 2
		if ( oPol.m_anX [ 0 ] != oPol.m_anX [ 2 ] ||
			  oPol.m_anY [ 0 ] != oPol.m_anY [ 2 ] ||
			  oPol.m_anZ [ 0 ] != oPol.m_anZ [ 2 ] ) {

			nP2X = oPol.m_anX [ 2 ] - oPol.m_anX [ 0 ];
			nP2Y = oPol.m_anY [ 2 ] - oPol.m_anY [ 0 ];
			nP2Z = oPol.m_anZ [ 2 ] - oPol.m_anZ [ 0 ];
		}
		// se demais vertices estavam sobrepostos nao formando vetores: 0 e 3
		else {

			nP2X = oPol.m_anX [ 3 ] - oPol.m_anX [ 0 ];
			nP2Y = oPol.m_anY [ 3 ] - oPol.m_anY [ 0 ];
			nP2Z = oPol.m_anZ [ 3 ] - oPol.m_anZ [ 0 ];
		}
	}
	else {

			nP1X = oPol.m_anX [ 2 ] - oPol.m_anX [ 0 ];
			nP1Y = oPol.m_anY [ 2 ] - oPol.m_anY [ 0 ];
			nP1Z = oPol.m_anZ [ 2 ] - oPol.m_anZ [ 0 ];

			nP2X = oPol.m_anX [ 3 ] - oPol.m_anX [ 0 ];
			nP2Y = oPol.m_anY [ 3 ] - oPol.m_anY [ 0 ];
			nP2Z = oPol.m_anZ [ 3 ] - oPol.m_anZ [ 0 ];
   }
   // calcula vetor normal aos 2 vetores obtidos
   INT16
      nNX = nP1Y * nP2Z - nP1Z * nP2Y,
      nNY = nP1Z * nP2X - nP1X * nP2Z,
      nNZ = nP1X * nP2Y - nP1Y * nP2X;

	// calcula o angulo entre os 2 vetores: iluminacao e normal
   // define o produto escalar entre os dois vetores (iluminacao e normal)
	float
		fProdEsc = nNX * m_nIX + nNY * m_nIY + nNZ * m_nIZ;

   // define o angulo (entre 0 e NANGS2 - 1)
   float
   	fV = (float) fProdEsc / ( MODULO ( nNX, nNY, nNZ ) * m_fModI );
   // verifica erros nos calculos
   if ( fV > 1 )
   	fV = 1;
   else
   	if ( fV < -1 )
      	fV = -1;
	INT16
		nAngulo = (INT16) ( acos ( fV ) * NANGS2 / Pi );

   // ajusta angulo para intervalo de 0 a NANGS / 4     (se angulo maior que
   // Porque:   VV   (iluminacao)                        NANGS2 calcula
   //           \/   (4 paredes)                         complementar)
   //           /\  todas devem receber a mesma iluminacao (sem sombras)
   if ( nAngulo > NANGS4 )
   	nAngulo = NANGS4 - nAngulo + NANGS4;

	// define cor fazendo regra de tres entre angulo e faixa cores utilizaveis
	return ( MINCOR + ( NUMCOR - 1 ) * nAngulo / NANGS4 );
}



/* Define uma cor de acordo com o valor Z do ponto medio do poligono.
return cor (entre MAXCOR e MINCOR) conforme FIMV. Quanto mais longe
(maior Z) mais proximo o valor a MINCOR.
**********************************************************/
INT16 CIlum::mRetCorZ ( CPol &oPol )
{
   // verifica se o objeto estah na area visivel
   INT16
   	nZ;
   if ( ( nZ = oPol.mRetCentroZ () ) < 0 )
   	return MAXCOR;
   else
   	if ( nZ > FIMIZ )
      	return MINCOR;
      else
			return ( MINCOR + (INT16) ( (float) - NUMCOR / FIMIZ * nZ + NUMCOR ) );
}
